// -*- mode:c++ -*-
//
// Header file clamp_support.hh
//
// Copyright (c) 2002, 2003 Raoul M. Gough
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// History
// =======
// 2002/ 2/ 7	rmg	File creation
//
// version: $Id: clamp_support.hh,v 1.14 2003/09/30 22:06:22 Raoul Exp $
//

#ifndef clamp_support_rmg_20020207_included
#define clamp_support_rmg_20020207_included

#include <string>
#include <list>
#include <memory>
#include <iosfwd>
#include "boost/smart_ptr.hpp"

namespace clamp
{
  typedef std::string stringT;

  struct paramT
  {
    boost::shared_ptr<stringT> type;
    boost::shared_ptr<stringT> name;
    boost::shared_ptr<stringT> initialiser;
    int ctx_param_num;
    bool byReference;

    paramT (stringT *t, stringT *n, stringT *i, int ctx = 0, bool ref = false)
      : type (t)
      , name (n)
      , initialiser (i)
      , ctx_param_num (ctx)
      , byReference (ref)
    {
    }
  };

  typedef std::list<boost::shared_ptr<paramT> > paramListT;

  struct lambdaT
  {
    std::auto_ptr<stringT> ret_type;
    std::auto_ptr<paramListT> context;
    std::auto_ptr<paramListT> params;
    std::auto_ptr<stringT> qual_str;
    bool explicitContext;
    bool onHeap;

    lambdaT ();
  };

  lambdaT *makelambda (lambdaT *lamp
		       , stringT *typestr
		       , paramListT *ctx
		       , paramListT *params
		       , stringT *qualstr
		       , bool onHeap = false);

  //
  // Convert an implicit-context lambda into an explicit-context version
  // by shifting the misplaced params into the correct context slot
  // and adding the given parameters
  //
  lambdaT *shiftlambda (lambdaT *lamp
			, paramListT *params);

  paramListT *makeplist (stringT *type, stringT *name, stringT *init);

  paramListT *appendplist (paramListT *front, paramListT *back);

  // Check if the plist contains any initialisers
  bool plistInitialisers (const paramListT *);

  stringT *makestr (const char *cstr);
  stringT *makestr (const char *cstr, int length);

  stringT *appendstr (stringT *, stringT *);
  stringT *appendstr (stringT *, stringT *, stringT *);
  stringT *appendstr (stringT *, stringT *, stringT *, stringT *);
  stringT *appendstr (stringT *, stringT *, stringT *, stringT *, stringT *);

  /*
  template<typename T>
  std::ostream &operator<< (std::ostream &strm, const T *tptr)
  {
    if (tptr)
      {
	return strm << *tptr;
      }

    else
      {
	return strm << "(null)";
      }
  }
  */

  std::ostream &operator<< (std::ostream &, const lambdaT &);
  std::ostream &operator<< (std::ostream &, const paramListT &);
}

#endif // clamp_support_rmg_20020207_included
