#include "error.hh"

#include <cstring>

/* This table of error messages comes almost directly
 * from QuickBASIC. Most of the messages are obsolete,
 * but they're maintained for compatibility with older
 * BASIC versions such as GW-BASIC.
 * (!) indicates that the message does not exist in QB.
 */
static const char* const ErrorMessages[] =
{
    "NEXT without FOR", //1,
    "Syntax error", //2,
    "RETURN without GOSUB", //3,
    "Out of DATA", //4,
    "Illegal function call", //5,
    "Overflow", //6,
    "Out of memory", //7,
    "Label not defined", //8,
    "Subscript out of range", //9,
    "Duplicate definition", //10,
    "Division by zero", //11,
    "Illegal in direct mode", //12,
    "Type mismatch", //13,
    "Out of string space", //14,
    "String too long", // 15 (!)
    "String formula too complex", //16,
    "Cannot continue", //17,
    "Function not defined", //18,
    "No RESUME", //19,
    "RESUME without error", //20,
    "Unprintable error", // 21
    "Missing operand", // 22 (!)
    "Line buffer overflow", // 23 (!)
    "Device timeout", //24,
    "Device fault", //25,
    "FOR without NEXT", //26,
    "Out of paper", //27,
    0,
    "WHILE without WEND", //29,
    "WEND without WHILE", //30,
    0,
    0,
    "Duplicate label", //33,
    0,
    "Subprogram not defined", //35,
    "Subprogram already in use", //36 (!),
    "Argument-count mismatch", //37,
    "Array not defined", //38,
    "CASE ELSE expected", //39,
    "Variable required", //40,
    0,0,0,0,
    0,0,0,0,
    0,
    "FIELD overflow", //50,
    "Internal error", //51,
    "Bad file name or number", //52,
    "File not found", //53,
    "Bad file mode", //54,
    "File already open", //55,
    "FIELD statement active", //56,
    "Device I/O error", //57,
    "File already exists", //58,
    "Bad record length", //59,
    0,
    "Disk full", //61,
    "Input past end of file", //62,
    "Bad record number", //63,
    "Bad file name", //64,
    0,
    "Direct statement in file", //66,
    "Too many files", //67,
    "Device unavailable", //68,
    "Communication-buffer overflow", //69,
    "Permission denied", //70,
    "Disk not ready", //71,
    "Disk-media error", //72,
    "Advanced feature unavailable", //73,
    "Rename across disks", //74,
    "Path/File access error", //75,
    "Path not found", //76,
    "Deadlock" //77,
};

std::vector<jmp_buf_wrapper> ErrorFrameStack;
jmp_buf_wrapper ErrorFrame;
void* ErrorHandler = 0;
void* ErrorReturnAddress = 0;
int B_ERR_VALUE;
const B_ERROR_LOCATION_TYPE* B_ERROR_LOCATION = 0;
const B_ERROR_LOCATION_TYPE* B_CURRENT_LOCATION = 0;

void B_RESUME_FAILTEST()
{
    if(!B_ERROR_LOCATION || ErrorFrameStack.empty())
    {
        B_ERROR(20);/* RESUME WITHOUT ERROR*/
    }
}

/* RESUME and RESUMENEXT may return into a subprogram, but also to main program */
void B_RESUME()
{
    B_RESUME_FAILTEST();
    
    ErrorReturnAddress = B_ERROR_LOCATION->resumerepeat;
    B_ERROR_LOCATION = 0; B_ERR_VALUE = 0; // clear error status
    longjmp(ErrorFrame, 1);
}

void B_RESUME_NEXT()
{
    B_RESUME_FAILTEST();

    ErrorReturnAddress = B_ERROR_LOCATION->resumenext;
    B_ERROR_LOCATION = 0; B_ERR_VALUE = 0; // clear error status
    longjmp(ErrorFrame, 1);
}

void B_RESUME_TO_do(void* label)
{
    B_RESUME_FAILTEST();
    
    /* Rewind error frame stack until at main */
    while(ErrorFrameStack.size() > 1) B_ERR_POP();
    
    ErrorReturnAddress = label;
    B_ERROR_LOCATION = 0; B_ERR_VALUE = 0; // clear error status
    longjmp(ErrorFrame, 1);
}

void B_ERR_PUSH()
{
    ErrorFrameStack.push_back(ErrorFrame);
}

void B_ERR_POP()
{
    ErrorFrame = ErrorFrameStack.back();
    ErrorFrameStack.pop_back();
}

void B_ERROR(BAS_INT code)
{
    if(B_ERROR_LOCATION || !ErrorHandler)
    {
        const BAS_INT NumMessages = sizeof(ErrorMessages) / sizeof(*ErrorMessages);
        int print_code = code;
        if(code < 1
        || code > NumMessages
        || !ErrorMessages[code-1]) print_code = 21; // Unprintable error
        
        static const B_ERROR_LOCATION_TYPE dummy_err = {0,0, 0,0,"<unknown>"};
        const B_ERROR_LOCATION_TYPE* e = B_ERROR_LOCATION ? B_ERROR_LOCATION : &dummy_err;
        
        fprintf(stderr, "%s (error %d) at %s:%d:%d\n",
            ErrorMessages[print_code-1], (int)code,
            e->file, e->line, e->col);
        exit(code);
    }
    
    B_ERROR_LOCATION = B_CURRENT_LOCATION;
    B_ERR_VALUE = code;
    
    goto *ErrorHandler;
}
