#ifndef bqtQBlibErrorHH
#define bqtQBlibErrorHH

#include <qblib/types.hh>

#include <setjmp.h>
#include <cstdio>
#include <vector>

class jmp_buf_wrapper
{
private:
    jmp_buf buf;
public:
    operator jmp_buf& () { return buf; }
};

extern std::vector<jmp_buf_wrapper> ErrorFrameStack;
extern jmp_buf_wrapper ErrorFrame;
extern void* ErrorReturnAddress;
extern void* ErrorHandler;

typedef struct{void* resumerepeat;
               void* resumenext;
               int line, col;
               const char *file;
              } B_ERROR_LOCATION_TYPE;
extern int B_ERR_VALUE;
extern const B_ERROR_LOCATION_TYPE *B_CURRENT_LOCATION;
extern const B_ERROR_LOCATION_TYPE *B_ERROR_LOCATION;

#define B_ERR_FRAME_BEGIN() \
    do { __label__ FrameBegin; \
         B_ERR_PUSH(); \
         if(unlikely(setjmp(ErrorFrame))) goto *ErrorReturnAddress; \
         static const B_ERROR_LOCATION_TYPE err = { &&FrameBegin, &&FrameBegin,0,0,0 }; \
         B_CURRENT_LOCATION = &err; \
         FrameBegin: ; \
       } while(0); \
    const struct AutoErrPop { ~AutoErrPop() { B_ERR_POP(); } } AutoErrPopper = { }

#define B_ERR_SETPOS(nextlabel, line, col, filename) \
    do { __label__ repeat_loc; \
         static const B_ERROR_LOCATION_TYPE err = \
            { &&repeat_loc, &&nextlabel, line,col,filename }; \
         B_CURRENT_LOCATION = &err; repeat_loc: ; \
       } while(0)

void B_RESUME_FAILTEST();
void B_ERR_PUSH();
void B_ERR_POP();
void B_RESUME();
void B_RESUME_NEXT();
void B_RESUME_TO_do(void* label);
void B_ERROR(BAS_INT code);

inline int B_FuncERR() { return B_ERR_VALUE; }
inline int B_FuncERL() { return B_ERROR_LOCATION ? B_ERROR_LOCATION->line : 0; }

/* RESUME_TO always goes into main program */
#define B_RESUME_TO(label) B_RESUME_TO_do(&&label)

#define ON_ERROR_GOTO(label) { if(likely(!HandlingError)) ErrorHandler=&&label; }
#define ON_ERROR_GOTO_NOWHERE() { ErrorHandler=0; }

#endif
